import os
from pkg_resources import resource_filename

import kivy
from kivy.app import App
from kivy.logger import Logger
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.slider import Slider
from kivy.uix.label import Label
from kivy.graphics import Color, Rectangle, PushMatrix, PopMatrix, Scale
from kivy.graphics.texture import Texture
from kivy.uix.widget import Widget
from kivy.metrics import sp
from kivy.uix.image import Image, AsyncImage
from kivy.uix.behaviors import ButtonBehavior
from kivy.properties import ObjectProperty, StringProperty, NumericProperty
from kivy.uix.behaviors import ButtonBehavior, FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.textinput import TextInput
from kivy.uix.popup import Popup
from kivy.properties import BooleanProperty
from kivy.config import Config
from kivy.utils import get_color_from_hex

# make sure we are on updated version of kivy
kivy.require('1.10.0')

normalfont = resource_filename(
        __name__,
        os.path.join('resources/fonts', 'DejaVuSans.ttf'))
boldfont = resource_filename(
        __name__,
        os.path.join('resources/fonts', 'DejaVuSans-Bold.ttf'))
fontawesomefont = resource_filename(
        __name__,
        os.path.join('resources/fonts', 'FontAwesome.ttf'))
buttonnormal = resource_filename(
        __name__,
        os.path.join('resources/images', 'button-normal.png'))
buttondown = resource_filename(
        __name__,
        os.path.join('resources/images', 'button-down.png'))
clearimage = resource_filename(
        __name__,
        os.path.join('resources/images', 'clear.png'))
backdrop = resource_filename(
        __name__,
        os.path.join('resources/images', 'backdrop.png'))
listbackdrop = resource_filename(
        __name__,
        os.path.join('resources/images', 'list-backdrop.png'))
trackslidercursor = resource_filename(
        __name__,
        os.path.join('resources/images', 'track-slider-cursor.png'))
ratingstars = [
        u"\uf006\uf006\uf006\uf006\uf006",
        u"\uf123\uf006\uf006\uf006\uf006",
        u"\uf005\uf006\uf006\uf006\uf006",
        u"\uf005\uf123\uf006\uf006\uf006",
        u"\uf005\uf005\uf006\uf006\uf006",
        u"\uf005\uf005\uf123\uf006\uf006",
        u"\uf005\uf005\uf005\uf006\uf006",
        u"\uf005\uf005\uf005\uf123\uf006",
        u"\uf005\uf005\uf005\uf005\uf006",
        u"\uf005\uf005\uf005\uf005\uf123",
        u"\uf005\uf005\uf005\uf005\uf005",
        u"\uf29c"]


class BaseLabel(Label):
    """Base label class to ensure consistency."""
    pass


class OutlineLabel(BaseLabel):
    """Label with an outline."""

    def __init__(self, *args, **kwargs):
        a = App.get_running_app()
        otype = int(a.config.get('colors','outlinetype'))
        w = int(a.config.get('colors','outlinewidth'))
        if otype == 1:
            self.outline_width = sp(w)
            (c1,c2,c3,c4) = get_color_from_hex(
                    a.config.get('colors','textoutline'))
            self.outline_color = (c1,c2,c3)
        elif otype == 2:
            self.bind(texture_size=self.draw_outline)
        super(OutlineLabel, self).__init__(*args, **kwargs)

    def draw_outline(self, *args, **kwargs):
        a = App.get_running_app()
        (c1,c2,c3,c4) = get_color_from_hex(
                a.config.get('colors','textoutline'))
        w = int(a.config.get('colors','outlinewidth'))
        x = int(self.center_x - self.texture_size[0] / 2.0)
        y = int(self.center_y - self.texture_size[1] / 2.0)
        # texture = Texture.create(size=self.texture_size,colorfmt='rgba')
        # widget = Widget()
        with self.canvas.before:
            for lw in range(w,0,-1):
                nw = self.texture_size[0] + lw*2
                nh = self.texture_size[1] + lw*2
                sw = nw / self.texture_size[0]
                sh = nh / self.texture_size[1]
                alpha = (1.0 / w) * (w-lw+1)
                Color(c1,c2,c3,alpha)
                Rectangle(pos=[x,y],size=self.texture_size,
                        texture=self.texture)
                PushMatrix()
                Scale(origin=(self.center_x,self.center_y),
                        x=sw, y=sh)
                PopMatrix()
        # reset the color, otherwise the next widget will be tinted
        Color(1.,1.,1.)
        # self.unbind(texture_size=self.draw_outline)
        # self.texture=widget.texture
        # self.bind(texture_size=self.draw_outline)


class ShadowLabel(BaseLabel):
    """Label with a drop shadow."""

    def __init__(self, *args, **kwargs):
        super(ShadowLabel, self).__init__(*args, **kwargs)
        # this was the only way i could get it to draw the correct size (?)
        self.bind(texture_size=self.draw_shadow)

    def draw_shadow(self, *args, **kwargs):
        a = App.get_running_app()
        o = int(a.config.get('colors','textshadowsize'))
        (c1,c2,c3,c4) = get_color_from_hex(a.config.get('colors','textshadow'))
        x = int(self.center_x - self.texture_size[0] / 2.0)
        y = int(self.center_y - self.texture_size[1] / 2.0)
        with self.canvas.before:
            # check if the texture is actually on the screen (?)
            if x > 0 and y > 0:
                for i in range(o):
                    alpha = (1.0 / (o+1)) * (o-i)
                    # Logger.debug("ShadowLabel: text "+self.text+" color "+format((c1,c2,c3,alpha))+" pos "+format([x+sp(i+1),y-sp(i-1)])+" size "+format(self.texture_size)+" texture "+format(self.texture))
                    Color(c1,c2,c3,alpha)
                    Rectangle(pos=[x+sp(i+1),y-sp(i+1)],
                            size=self.texture_size, texture=self.texture)
            # reset the color, otherwise the next widget will be tinted
            Color(1.,1.,1.)


class ScaledWidthLabel(BaseLabel):
    """Helps avoid a divide by zero exception when texture_size[0] = 0 (?)"""
    _internalscale = NumericProperty(None)
    _scale = NumericProperty(1.0)

    def __init__(self, *args, **kwargs):
        super(ScaledWidthLabel, self).__init__(*args, **kwargs)
        if len(self.text) < 50:
            self.font_size = '40sp'
        else:
            self.font_size = '20sp'
        self.do_resize()

    def do_resize(self):
        if self.texture_size[0] > 0 and self.texture_size[0] < self.width:
            self._internalscale = float(self.width) / self.texture_size[0]
        else:
            self._internalscale = 1.0

    def on_texture_size(self, obj, value):
        self.do_resize()


class ExtraSlider(Slider):
    """Class that implements some extra stuff on top of a standard slider."""

    def __init__(self, **kwargs):
        """Do normal init routine, but also register on_release event."""
        super(self.__class__, self).__init__(**kwargs)
        self.register_event_type('on_release')

    def on_release(self):
        """Override this with something you want this slider to do."""
        pass

    def on_touch_up(self, touch):
        """Check if slider is released, dispatch the on_release event if so."""
        released = super(self.__class__, self).on_touch_up(touch)
        if released:
            self.dispatch('on_release')
        return released


class CoverButton(Button):
    """Album cover that is pressable."""

    img = ObjectProperty(None)
    layout = ObjectProperty(None)

    def __init__(self, **kwargs):
        super(self.__class__, self).__init__(**kwargs)
        self.background_normal = clearimage
        self.background_down = clearimage
        self.font_name = boldfont
        with self.canvas.before:
            Rectangle(texture=self.img.texture,
                      pos=self.layout.pos,
                      size=self.layout.size)


class ArtistRecycleBoxLayout(
        FocusBehavior, LayoutSelectionBehavior, RecycleBoxLayout):
    ''' Adds selection and focus behaviour to the view. '''


class ArtistRow(RecycleDataViewBehavior, BoxLayout):
    ''' Add selection support to the Label '''
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    def refresh_view_attrs(self, rv, index, data):
        ''' Catch and handle the view changes '''
        self.index = index
        return super(ArtistRow, self).refresh_view_attrs(
            rv, index, data)

    def on_touch_down(self, touch):
        ''' Add selection on touch down '''
        if super(ArtistRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        ''' Respond to the selection of items in the view. '''
        self.selected = is_selected
        if is_selected:
            App.get_running_app().root.selected_row = index


class RatingPopup(Popup):
    rating_set = ObjectProperty(None)
    song = StringProperty(None)
    index = NumericProperty(None)
    robject = ObjectProperty(None)


class ScreenSwitcher(BoxLayout):
    manager = ObjectProperty(None)
    currentscreen = StringProperty('nowplaying')




import os
import sys
import mutagen
import io
import random
import socket
import re
import json
from functools import partial

from PIL import Image as PImage
import kivy
from kivy.app import App
from kivy.logger import Logger
from kivy.uix.screenmanager import Screen
from kivy.graphics import Rectangle, Color
from kivy.core.image import Image as CoreImage
from kivy.clock import Clock
from kivy.uix.button import Button
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.image import Image
from kivy.uix.popup import Popup
from kivy.properties import ObjectProperty
from kivy.core.window import Window
from kivy.factory import Factory
from kivy.metrics import sp
from twisted.internet.defer import Deferred, DeferredList

from kmpc.extra import KmpcHelpers
from kmpc.widgets import (CoverButton, clearimage, ratingstars,
                          normalfont, fontawesomefont, backdrop)

# make sure we are on updated version of kivy
kivy.require('1.10.0')

Helpers = KmpcHelpers()

class NowPlayingScreen(Screen):
    """The Now Playing screen."""

    def __init__(self, *args, **kwargs):
        super(NowPlayingScreen, self).__init__(*args, **kwargs)
        self.accessoryPopup = Factory.AccessoryPopup()
        self.settingsPopup = Factory.SettingsPopup()
        # create the once-per-second update of the track slider
        self.track_slider_task = Clock.schedule_interval(
                self.update_track_slider, 1)
        self.track_slider_task.cancel()

    def enter_screen(self, *args):
        ss = self.ids.ss
        ss.ids.nowplayingbutton.state = 'down'
        ss.ids.nowplayingbutton.disabled = True
        app = App.get_running_app().root
        if app:
            (app.mpdconnection.protocol.status().
                addCallback(self.update_mpd_status).
                addErrback(app.mpdconnection.handle_mpd_error))

    def leave_screen(self, *args):
        # pause the track slider task
        self.track_slider_task.cancel()

    def settings_popup(self):
        self.settingsPopup.open()
        app = App.get_running_app().root
        # get the host's IP address and display it
        iptext = "IP Address: " + format(self.get_ip())
        self.settingsPopup.ids.ip_label.text = iptext
        (app.mpdconnection.protocol.status().
            addCallback(partial(self.update_mixers, self.settingsPopup)).
            addErrback(app.mpdconnection.handle_mpd_error))
        (app.mpdconnection.protocol.replay_gain_status().
            addCallback(partial(self.update_replaygain, self.settingsPopup)).
            addErrback(app.mpdconnection.handle_mpd_error))

    def update_replaygain(self, p, result):
        rg = str(result)
        p.ids['rg_'+rg].state = 'down'

    def update_mixers(self, p, result):
        # set up the crossfade slider
        if 'xfade' in result:
            v = int(result['xfade'])
        else:
            v = 0
        p.ids.crossfade_slider.value = v
        # set up the mixrampdb slider
        if 'mixrampdb' in result:
            v = round(float(result['mixrampdb']), 6)
        else:
            v = 0.0
        p.ids.mixrampdb_slider.value = float(str(v)[1:])
        # set up the mixrampdelay slider
        if 'mixrampdelay' in result:
            v = round(float(result['mixrampdelay']), 6)
        else:
            v = 0.0
        p.ids.mixrampdelay_slider.value = v

    def get_ip(self):
        """Method that tries to get the local IP address, and returns localhost
        if there isn't one."""
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except Exception:
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP

    def handle_mpd_message(self, result):
        """Callback for mpd 'kmpc' messages."""
        # result is an iterator although it only generally has one member
        for m in result:
            message = m['message']
            Logger.info("Application: MPD Message: "+message)
            if message == 'AccessoryOff':
                # generated by mausberry circuit, accessory power is off
                self.accessoryPopup.open()
            elif message == 'AccessoryOn':
                # generated by mausberry circuit, accessory power is on
                self.accessoryPopup.dismiss()

    def current_track_slider_release(self):
        """Callback when track slider is released."""
        # get the value of the track slider
        curpos = int(self.ids.current_track_slider.value)
        Logger.debug("Application: touch up on track slider at "+str(curpos))
        app = App.get_running_app().root
        # seek to that position in the track
        app.mpdconnection.protocol.seekcur(
                str(curpos)).addErrback(app.mpdconnection.handle_mpd_error)

    def current_track_slider_down(self):
        """Callback when track slider is pressed."""
        Logger.debug("Application: touch down on track slider")
        # cancel the once-per-second update while finger is down so it doesn't
        # move out from under it
        self.track_slider_task.cancel()

    def update_track_slider(self, dt):
        """Increment the slider by 1 second every second."""
        Logger.debug("Application: update_track_slider")
        curpos = int(self.ids.current_track_slider.value) + 1
        self.ids.current_track_slider.value = curpos

    def stop_zero_stuff(self):
        """Clear the screen and variables if playback is fully stopped."""
        app = App.get_running_app().root
        self.ids.current_track_slider.value = 0
        self.ids.current_track_slider.max = 0
        self.ids.current_queue_track_number_label.text = ''
        self.ids.next_song_artist_label.text = ''
        app.currfile = None
        app.currsong = None
        app.nextsong = None
        self.ids.song_star_button.text = ''
        self.ids.song_star_button.disabled = True
        self.ids.album_cover_layout.clear_widgets()
        self.ids.trackinfo.clear_widgets()
        lbl = Factory.BaseLabel(text="Playback Stopped")
        self.ids.trackinfo.add_widget(lbl)
        self.ids.player.canvas.before.clear()
        self.ids.releasetypelabel.text = ''
        self.ids.yearlabel.text = ''
        self.ids.remasterlabel.text = ''

    def update_mpd_status(self, result):
        """Callback when mpd status changes."""
        Logger.debug('NowPlaying: update_mpd_status()')
        # probably state is the only one necessary, but hey
        app = App.get_running_app().root
        app.mpd_status['state'] = result['state']
        app.mpd_status['repeat'] = result['repeat']
        app.mpd_status['single'] = result['single']
        app.mpd_status['random'] = result['random']
        app.mpd_status['consume'] = result['consume']
        if app.mpd_status['state'] == 'stop':
            # if stopped, there are no current or next songs
            app.currsong = None
            app.nextsong = None
            Logger.debug("update_mpd_status: Nothing playing")
        else:
            # save current song, this is a 0-based index into the queue
            app.currsong = result['song']
            # ask mpd for current song data
            (app.mpdconnection.protocol.currentsong().
                addCallback(partial(self.update_mpd_currentsong, False)).
                addErrback(app.mpdconnection.handle_mpd_error))
            # save next song, this is a 0-based index into the queue
            if 'nextsong' in result:
                app.nextsong = result['nextsong']
                # ask mpd for next song data
                (app.mpdconnection.protocol.playlistinfo(app.nextsong).
                    addCallback(self.update_mpd_nextsong).
                    addErrback(app.mpdconnection.handle_mpd_error))
            else:
                app.nextsong = None
        stflags = ['normal', 'down']
        # check various flag states
        self.ids.repeat_button.state = stflags[
                int(app.mpd_status['repeat'])]
        self.ids.single_button.state = stflags[
                int(app.mpd_status['single'])]
        self.ids.random_button.state = stflags[
                int(app.mpd_status['random'])]
        self.ids.consume_button.state = stflags[
                int(app.mpd_status['consume'])]
        if (app.mpd_status['state'] == 'pause' or
                app.mpd_status['state'] == 'stop'):
            # play/pause button should be a play button that is unpressed
            self.ids.play_button.state = 'normal'
            self.ids.play_button.text = u"\uf04b"
            self.track_slider_task.cancel()
        else:
            # play/pause button should be a pause button that is pressed
            self.ids.play_button.state = 'down'
            self.ids.play_button.text = u"\uf04c"
        if app.mpd_status['state'] == 'stop':
            # zero everything out if we are stopped
            self.stop_zero_stuff()
        else:
            # mpd returns {elapsed seconds}:{total seconds}, the following
            # splits each to minute:second
            c, t = result['time'].split(":")
            # set the max slider value to the total seconds
            self.ids.current_track_slider.max = int(t)
            # set the current slider value to the current seconds
            app.mpd_status['curpos'] = int(c)
            self.ids.current_track_slider.value = int(c)
            if app.mpd_status['state'] != 'pause':
                self.track_slider_task()
            # throws an exception if i don't do this
            a = int(result['song'])+1
            b = int(result['playlistlength'])
            self.ids.current_queue_track_number_label.text = \
                "%d of %d" % (a, b)
        # update the queue tab with status results so that current track
        # will be highlighted
        app.queue.update_mpd_status(result)

    def change_artist_image(self, img, al_path, instance):
        """Called when you click on an artist logo, changes it to another at
        random."""
        Logger.debug("change_artist_image: (current path is "+img.source+")")
        img_path = img.source
        if len(os.listdir(al_path)) > 1:
            while img_path == img.source:
                img_path = os.path.join(
                        al_path,
                        random.choice(os.listdir(al_path)))
            if os.path.isfile(img_path):
                img.source = img_path
            Logger.debug("change_artist_image: (new path is "+img.source+")")
        else:
            Logger.debug("change_artist_image: no other choices for logo")

    def get_cache(self, ctype, uuid):
        """Returns data object for an artist, recording, or release from
        cache."""
        # pull the mb-cache folder from ini file
        cachepath = App.get_running_app().root.config.get('paths', 'cachepath')
        mbpath = os.path.join(cachepath, ctype, Helpers.upath(uuid),
                uuid+'.json')
        Logger.debug("Application: reading cache data from "+mbpath)
        data = {}
        try:
            with open(mbpath) as json_file:
                data = json.load(json_file)
        except:
            Logger.debug("Application: cache file could not be read")
        return data

    def get_artist_image(self, artist_id):
        img = None
        cachepath = App.get_running_app().root.config.get('paths', 'cachepath')
        try:
            # look in the 'logo' subfolder of the artistid path
            #al_path = os.path.join(fa_path, mb_aid, 'logo')
            al_path = os.path.join(cachepath, 'artist', Helpers.upath(artist_id), 'logo')
            # pick one at random
            img_path = os.path.join(
                    al_path,
                    random.choice(os.listdir(al_path)))
            if os.path.isfile(img_path):
                # create an image button out of the logo so you can
                # press it
                img = Factory.ImageButton(
                        source=os.path.join(al_path, img_path),
                        allow_stretch=True,
                        color=(1, 1, 1, 0.65))
                img.bind(on_press=partial(self.change_artist_image,
                                          img, al_path))
        except OSError:
            Logger.debug(
                    "update_mpd_currentsong: No logos for artist "
                    + artist_id)
        except Exception as e:
            Logger.exception("update_mpd_currentsong: "+format(e))
        return img

    def get_cover_from_file(self, p, res):
        # try to pull the cover art from the file
        img = None
        if os.path.isfile(p):
            Logger.debug('get_cover_from_file: found good file at path ' + p)
            # load up the file to read the tags
            f = mutagen.File(p)
            cimg = None
            data = None
            # try to get mp3 cover, if this throws an exception it's
            # not an mp3 or it doesn't have a cover
            try:
                pframes = f.tags.getall("APIC")
                # id3v2 can store any number of image frames, we just
                # want the first one
                for frame in pframes:
                    ext = 'img'
                    # figure out the file type
                    if (frame.mime.endswith('jpeg') or
                            frame.mime.endswith('jpg')):
                        ext = 'jpg'
                    elif frame.mime.endswith('png'):
                        ext = 'png'
                    elif frame.mime.endswith('bmp'):
                        ext = 'bmp'
                    elif frame.mime.endswith('gif'):
                        ext = 'gif'
                    # pull the raw image data into a variable
                    data = io.BytesIO(bytearray(frame.data))
                    break
            except AttributeError:
                pass
            # try to get mp4 cover
            if 'covr' in f.keys():
                # figure out the file type
                if f['covr'][0].imageformat == \
                        mutagen.mp4.MP4Cover.FORMAT_PNG:
                    ext = 'png'
                else:
                    ext = 'jpg'
                # pull the raw image data into a variable
                data = io.BytesIO(bytearray(f['covr'][0]))
            if data:
                # if we got image data, load it as a kivy.core.image
                # filter through PIL first to resize it if it is too
                # large for a texture
                pimg = PImage.open(data)
                (w, h) = pimg.size
                (mw, mh) = Window.size
                if w > mw or h > mh:
                    Logger.debug('get_cover_from_file: '
                            + 'resizing cover image to maximum of '
                            + format(mh)+'x'
                            + format(mh))
                    pimg.thumbnail((mh, mh))
                    data2 = io.BytesIO()
                    pimg.convert('RGB').save(data2, 'PNG')
                    data2.seek(0)
                    cimg = CoreImage(data2, ext='png')
                else:
                    Logger.debug(
                            'get_cover_from_file: '
                            + 'pulling cover directly from tag')
                    data.seek(0)
                    cimg = CoreImage(data, ext=ext)
            if cimg:
                img = CoverButton(
                        img=cimg,
                        layout=self.ids.album_cover_layout,
                        halign='center')
            else:
                img = CoverButton(
                        img=CoreImage(clearimage),
                        layout=self.ids.album_cover_layout,
                        halign='center')
        else:
            # this should _probably_ never happen
            Logger.debug('NowPlaying: no file found at path '+p)
        return img

    def format_track_title(self, title):
        # check to see if song title has any data deliminated by ()
        # or []
        stitle = re.split('[\(\[\]\)]', title)
        # if the first item is empty, the title starts with () or
        # []
        if not stitle[0]:
            stitle.pop(0)
            stitle[0] = '('+stitle[0]+')'
        ftitle = filter(None, stitle)
        Logger.debug('TITLE: '+format(stitle))
        if len(ftitle) > 1:
            title_widget = BoxLayout(orientation='vertical',
                             padding_y='2sp')
            # split the title up and put the parentheses in smaller
            # text below
            title_widget.add_widget(Factory.InfoLargeLabel(text=stitle[0]))
            l2 = ""
            for i, v in enumerate(stitle):
                if i > 0 and v.strip():
                    if l2:
                        l2 += ' '
                    l2 += '('+v.strip()+')'
            title_widget.add_widget(Factory.InfoLargeLabel(text=l2,
                    _scale=0.5))
        else:
            title_widget = Factory.InfoLargeLabel(text=title)
        return title_widget

    def format_album_title(self, album):
        # check to see if album title is a split (has a ' / ' in
        # the middle)
        talbum = album.split(' / ')
        Logger.debug('ALBUM1: '+format(talbum))
        album_widget = BoxLayout(orientation='horizontal', padding_x='2sp')
        if len(talbum) > 1:
            Logger.debug('ALBUM1: album is a split')
            split = True
        else:
            split = False
        for j, a in enumerate(talbum):
            if j > 0:
                album_widget.add_widget(Factory.InfoLargeLabel(
                        font_size='50sp',
                        text=u'\u2571',
                        size_hint_x=None,
                        font_name=normalfont))
            # check to see if album title has any data deliminated
            # by () or []
            salbum = re.split('[\(\[\]\)]', a)
            Logger.debug('ALBUM2: '+format(salbum))
            if len(salbum) > 1:
                bl2 = BoxLayout(
                        orientation='vertical',
                        padding_y='2sp')
                # split the album up and put the parentheses in
                # smaller text below
                if split:
                    bl2.add_widget(Factory.InfoLargeLabel(text=salbum[0],
                            _scale=0.66))
                else:
                    bl2.add_widget(Factory.InfoLargeLabel(text=salbum[0]))
                l2 = ""
                for i, v in enumerate(salbum):
                    if i > 0 and v.strip():
                        if l2:
                            l2 += ' '
                        l2 += '('+v.strip()+')'
                bl2.add_widget(Factory.InfoLargeLabel(text=l2, _scale=0.5))
                album_widget.add_widget(bl2)
            else:
                if split:
                    album_widget.add_widget(Factory.InfoLargeLabel(text=a,
                            _scale=0.66))
                else:
                    album_widget.add_widget(Factory.InfoLargeLabel(text=a))
        return album_widget

    def create_artist_widget(self, recording_data, mpd_data):
        artist_ids = []
        if recording_data:
            Logger.debug("create_artist_widget: parsing cache data")
            # create an artist widget
            artist_bl = BoxLayout(size_hint=(1, 1),
                orientation='horizontal', spacing=0)
            # loop through artist credits
            f_flag = False
            for row in recording_data['recording']['artist-credit']:
                if isinstance(row, dict) and 'artist' in row:
                    # found artist block
                    artist_id = row['artist']['id']
                    artist_ids.append(artist_id)
                    artist_name = row['artist']['name']
                    artist_image = self.get_artist_image(artist_id)
                    if f_flag:
                        nbl = BoxLayout(orientation='vertical')
                        ftext = Factory.InfoSmallLabel(text="featuring")
                        nbl.add_widget(ftext)
                    else:
                        nbl = artist_bl
                    if artist_image:
                        nbl.add_widget(artist_image)
                    else:
                        atext = Factory.InfoLargeLabel(text=artist_name)
                        nbl.add_widget(atext)
                    if f_flag:
                        artist_bl.add_widget(nbl)
                        f_flag = False
                elif 'feat' in row.lower():
                    # found featuring string, set the flag
                    f_flag = True
                elif isinstance(row, basestring):
                    # found join characters
                    jtext = Factory.InfoSmallLabel(text=row, size_hint_x=None)
                    artist_bl.add_widget(jtext)
            return (artist_ids, artist_bl)
        else:
            Logger.debug("create_artist_widget: parsing tags")
            t = mpd_data['artist']
            artist_text = Factory.InfoLargeLabel(text=t)
            if 'musicbrainz_artistid' in mpd_data:
                artist_ids = mpd_data['musicbrainz_artistid'].split('/')
            return (artist_ids, artist_text)

    def create_track_widget(self, recording_data, mpd_data):
        if recording_data:
            t = recording_data['recording']['title']
        else:
            t = mpd_data['title']
        if App.get_running_app().root.config.getboolean('system',
                'advancedtitles'):
            title_text = self.format_track_title(t)
        else:
            title_text = Factory.InfoLargeLabel(text=t)
        return title_text

    def create_album_widget(self, release_data, mpd_data):
        if release_data:
            t = release_data['release']['title']
        else:
            t = mpd_data['album']
        if App.get_running_app().root.config.getboolean('system',
                'advancedtitles'):
            album_text = self.format_album_title(t)
        else:
            album_text = Factory.InfoLargeLabel(text=t)
        return album_text

    def update_mpd_currentsong(self, force, result):
        """Callback for mpd currentsong data."""
        Logger.debug('NowPlaying: update_mpd_currentsong()')
        app = App.get_running_app().root
        # this is so expensive screen updates only happen if the song has
        # changed since the last time this callback was called
        songchange = False
        # ti is the track info widget
        ti = self.ids.trackinfo
        # if result is undefined, there's not actually a song playing
        if result:
            # if class's current file doesn't match what mpd returns, the song
            # has changed
            if app.currfile != result['file']:
                songchange = True
            # update class's current file
            app.currfile = result['file']
            if songchange or force:
                # clear the track info widget
                ti.clear_widgets()
                # clear the album cover
                self.ids.album_cover_layout.clear_widgets()
                # get the stored star rating
                (app.mpdconnection.protocol.
                    sticker_get('song', app.currfile,
                            'rating').
                    addCallback(self.update_mpd_sticker_rating).
                    addErrback(self.handle_mpd_no_sticker))
                # clear the background of the player canvas so we can show the
                # artist background
                self.ids.player.canvas.before.clear()
                # figure out the full path of the file
                bp = app.config.get('paths', 'musicpath')
                # p is the absolute path
                p = os.path.join(bp, result['file'])
                # pull the mb-cache folder from ini file
                cachepath = app.config.get('paths', 'cachepath')
                # try to get recording data from cache
                if 'musicbrainz_trackid' in result:
                    recording_data = self.get_cache('recording',
                            result['musicbrainz_trackid'])
                # try to get release data from cache
                if 'musicbrainz_albumid' in result:
                    release_data = self.get_cache('release',
                            result['musicbrainz_albumid'])

                # not entirely convinced that the size_hints in the following 
                # widgets actually do any good
                # create an artist title widget and get a list of artist_ids
                (artist_ids, artist_widget) = self.create_artist_widget(
                        recording_data, result)
                artist_widget.size_hint_y = 1
                artist_widget.size_hint_min_y = sp(95)
                ti.add_widget(artist_widget)

                # create a track title widget
                title_text = self.create_track_widget(recording_data, result)
                title_text.size_hint_y = 1
                title_text.size_hint_max_y = sp(95)
                ti.add_widget(title_text)

                # create an album title widget
                album_text = self.create_album_widget(release_data, result)
                album_text.size_hint_y = 1
                album_text.size_hint_max_y = sp(95)
                ti.add_widget(album_text)

                # figure out release dates
                yy = {'year': None, 'oyear': None,
                        'rdate': None, 'ordate': None}
                if release_data:
                    rg = release_data['release']['release-group']
                    if ('date' not in release_data['release'] and
                            'first-release-date' in rg):
                        release_data['release']['date'] = rg['first-release-date']
                    if 'date' in release_data['release']:
                        yy['year'] = release_data['release']['date'][:4]
                        if len(release_data['release']['date']) > 4:
                            yy['rdate'] = release_data['release']['date']
                    if ('first-release-date' in rg
                            and (rg['first-release-date'] !=
                                    release_data['release']['date'])):
                        yy['oyear'] = rg['first-release-date'][:4]
                        if len(rg['first-release-date']) > 4:
                            yy['ordate'] = rg['first-release-date']
                else:
                    if 'date' in result:
                        yy['year'] = result['date'][:4]
                        if len(result['date']) > 4:
                            yy['rdate'] = result['date']
                    if ('originaldate' in result
                            and result['date'] != result['originaldate']):
                        yy['oyear'] = result['originaldate'][:4]
                        if len(result['originaldate']) > 4:
                            yy['ordate'] = result['originaldate']

                # parse the release type
                if release_data:
                    t = release_data['release']['release-group']['type']
                    rl = self.ids.releasetypelabel
                    rl.text = t
                    rl.font_size = '20sp'
                else:
                    self.ids.releasetypelabel.text = ''

                # pick an artistid at random for the background
                if artist_ids:
                    mb_aid = random.choice(artist_ids)
                    try:
                        # look in the 'artistbackground' subfolder of the artistid
                        # path
                        ab_path = os.path.join(cachepath, 'artist',
                                Helpers.upath(mb_aid), 'artistbackground')
                        # pick one at random
                        img_path = random.choice(os.listdir(ab_path))
                        # update the player background with the image
                        with self.ids.player.canvas.before:
                            Color(1, 1, 1)
                            Rectangle(
                                    source=os.path.join(ab_path, img_path),
                                    size=self.ids.player.size,
                                    pos=self.ids.player.pos)
                    except Exception:
                        # if we can't get an artistbackground image, just do
                        # nothing
                        pass

                # try to pull the cover art from the file
                callbacks = []
                d = Deferred()
                callbacks.append(d.addCallbacks(
                        partial(self.get_cover_from_file, p),
                        app.mpdconnection.handle_mpd_error))
                callbacks.append(d.addCallbacks(
                        partial(self.handle_get_cover,
                                release_data, result, yy),
                        app.mpdconnection.handle_mpd_error))
                dlist = DeferredList(callbacks, consumeErrors=False)
                d.callback('BEGIN')

                # format year labels
                if yy['year']:
                    self.ids.yearlabel.text = "[" + yy['year'] + "]"
                else:
                    self.ids.yearlabel.text = ""
                if (app.config.getboolean('system', 'originalyear')
                        and yy['oyear'] and yy['year'] != yy['oyear']):
                    self.ids.yearlabel.text = "["+yy['oyear']+"]"
                    self.ids.remasterlabel.text = "{"+yy['year']+"}"
                else:
                    self.ids.remasterlabel.text = ""

        else:
            # there's not a current song, so zero everything out
            self.stop_zero_stuff()

    def handle_get_cover(self, release_data, mpd_data, yy, img):
        # create cover art popup
        if img:
            app = App.get_running_app().root
            Logger.debug("handle_get_cover: got an image")
            self.ids.album_cover_layout.add_widget(img)
            # popup the cover large if you press it
            if release_data:
                Logger.debug("handle_get_cover: formatting title from cache")
                try:
                    t = release_data['release']['release-group']['type']
                    title = (t + ' "' + release_data['release']['title'] +
                            '" by ')
                    t = release_data['release']['artist-credit-phrase']
                    title += (t + ' released ' +
                            Helpers.country(
                                    release_data['release']['country']))
                    if yy['rdate']:
                        title += ' on ' + yy['rdate']
                    elif yy['year']:
                        title += ' in ' + yy['year']
                    if (app.config.getboolean('system', 'originalyear')
                            and yy['oyear']):
                        if yy['ordate']:
                            title += (' (originally released ' +
                                yy['ordate'] + ')')
                        else:
                            title += (' (originally released ' +
                                yy['oyear'] + ')')
                    if 'disambiguation' in release_data['release']:
                        t = release_data['release']['disambiguation']
                        title += ' [' + t + ']'
                except Exception as e:
                    Logger.error("handle_get_cover: "+format(e))
            else:
                Logger.debug("handle_get_cover: formatting title from tags")
                title = ('"' + mpd_data['album'] + '" by ' +
                        mpd_data['albumartist'] + ' released ')
                if yy['rdate']:
                    title += 'on ' + yy['rdate']
                elif yy['year']:
                    title += 'in ' + yy['year']
                if (app.config.getboolean('system', 'originalyear') and
                        yy['oyear']):
                    if yy['ordate']:
                        title += (' (originally released ' + yy['ordate']
                            + ')')
                    else:
                        title += (' (originally released ' + yy['oyear']
                            + ')')
            img.bind(on_press=partial(self.cover_popup, title))
        return True

    def update_mpd_sticker_rating(self, result):
        """Callback for song that has a rating in mpd."""
        Logger.debug('NowPlaying: update_mpd_sticker_rating')
        self.ids.song_star_button.disabled = False
        self.ids.song_star_button.text = ratingstars[int(result)]

    def handle_mpd_no_sticker(self, result):
        """Callback for song that has no rating in mpd."""
        Logger.debug('NowPlaying: handle_mpd_no_sticker')
        self.ids.song_star_button.disabled = False
        self.ids.song_star_button.text = ratingstars[11]

    def update_mpd_nextsong(self, result):
        """Callback for next song data from mpd."""
        Logger.debug('NowPlaying: update_mpd_nextsong()')
        # result is a list with one member
        for obj in result:
            # set the next song label
            self.ids.next_song_artist_label.text = \
                    'Up Next: '+obj['artist']+' - '+obj['title']

    def prev_pressed(self):
        """Callback for prev button pressed."""
        Logger.debug('Application: prev_pressed()')
        App.get_running_app().root.mpdconnection.protocol.previous()

    def play_pressed(self):
        """Callback for play/pause button pressed."""
        Logger.debug('Application: play_pressed()')
        app = App.get_running_app().root
        if app.mpd_status['state'] == 'play':
            # pause if playing
            app.mpdconnection.protocol.pause()
        else:
            # play if paused or stopped
            app.mpdconnection.protocol.play()

    def next_pressed(self):
        """Callback for next button pressed."""
        Logger.debug('Application: next_pressed()')
        App.get_running_app().root.mpdconnection.protocol.next()

    def repeat_pressed(self):
        """Callback for repeat button pressed."""
        Logger.debug('Application: repeat_pressed()')
        app = App.get_running_app().root
        # toggle on/off
        app.mpdconnection.protocol.repeat(
                str(1-int(app.mpd_status['repeat'])))

    def single_pressed(self):
        """Callback for single button pressed."""
        Logger.debug('Application: single_pressed()')
        app = App.get_running_app().root
        # toggle on/off
        app.mpdconnection.protocol.single(
                str(1-int(app.mpd_status['single'])))

    def random_pressed(self):
        """Callback for random button pressed."""
        Logger.debug('Application: random_pressed()')
        app = App.get_running_app().root
        # toggle on/off
        app.mpdconnection.protocol.random(
                str(1-int(app.mpd_status['random'])))

    def consume_pressed(self):
        """Callback for consume button pressed."""
        Logger.debug('Application: consume_pressed()')
        app = App.get_running_app().root
        # toggle on/off
        app.mpdconnection.protocol.consume(
                str(1-int(app.mpd_status['consume'])))

    def rating_popup(self, instance):
        """Popup for setting song rating."""
        Logger.debug('Application: rating_popup()')
        popup = Factory.RatingPopup(
                rating_set=self.rating_set,
                song=App.get_running_app().root.currfile)
        popup.open()

    def rating_set(self, song, rating, popup):
        """Method that sets a song's rating."""
        Logger.debug('Application: rating_set('+rating+')')
        app = App.get_running_app().root
        popup.dismiss()
        if rating:
            app.mpdconnection.protocol.sticker_set(
                    'song',
                    song,
                    'rating',
                    rating)
            app.mpdconnection.protocol.sendmessage("ratingchange",song)
        else:
            app.mpdconnection.protocol.sticker_delete(
                    'song',
                    song,
                    'rating')

    def cover_popup(self, title, instance):
        """Popup for showing a larger version of the album cover."""
        Logger.debug('Application: cover_popup()')
        popup = Popup(title=title,size_hint=(0.6,1))
        popup.add_widget(Image(texture=instance.img.texture))
        popup.open()

    def change_backlight(self, value):
        """Method that sets the backlight to a certain value."""
        app = App.get_running_app().root
        Logger.info('Application: '
                    + 'change_backlight('+str(value)
                    + ') rpienable = '
                    + app.config.get('system', 'rpienable'))
        # only if the ini file says it's ok
        if app.config.getboolean('system', 'rpienable'):
            import rpi_backlight as bl
            # set the brightness
            bl.set_brightness(int(value), smooth=True, duration=1)

    def settings_update(self):
        app = App.get_running_app().root
        (app.mpdconnection.protocol.currentsong().
            addCallback(partial(self.update_mpd_currentsong, True)).
            addErrback(app.mpdconnection.handle_mpd_error))




import os
import errno
import io
import pycountry
from PIL import Image as PImage

from twisted.internet.defer import Deferred, DeferredList


class KmpcHelpers(object):

    def formatsong(self, rec):
        """Method used by library browser to properly format a song row."""
        song = ''
        # check if there is more than one disc and display if so
        dd = rec['disc'].split('/')
        if len(dd) > 1:
            if int(dd[1]) > 1:
                song += '(Disc ' + '%02d' % int(dd[0]) + ') '
        # sometimes track numbers are like '01/05' (one of five), so drop that
        # second number
        tt = rec['track'].split('/')
        song += '%02d' % int(tt[0]) + ' '
        # if albumartist is different than track artist, display the track
        # artist
        if rec['artist'] != rec['albumartist']:
            song += rec['artist'] + ' - '
        # display the track title
        song += rec['title']
        return song

    def decodeFileName(self, name):
        """Method that tries to intelligently decode a filename to handle
        unicode weirdness."""
        if type(name) == str:
            try:
                name = name.decode('utf8')
            except Exception:
                name = name.decode('windows-1252')
        return name

    def removeEmptyFolders(self, path, removeRoot=True):
        """Function to remove empty folders"""
        if not os.path.isdir(path):
            return
        # remove empty subfolders
        files = os.listdir(path)
        if len(files):
            for f in files:
                fullpath = os.path.join(path, f)
                if os.path.isdir(fullpath):
                    self.removeEmptyFolders(fullpath)
        # if folder empty, delete it
        files = os.listdir(path)
        if len(files) == 0 and removeRoot:
            os.rmdir(path)

    def upath(self, uuid):
        """Function to expand uuid to full cache path."""
	p1 = uuid[0:2]
	p2 = uuid[2:4]
	p3 = uuid[4:6]
	p4 = uuid[6:8]
	return os.path.join(p1,p2,p3,p4,uuid)
    
    def expath(self, cachepath, uuid, idtype):
        return os.path.join(cachepath,idtype,self.upath(uuid))

    def jpath(self, cachepath, uuid, idtype):
        return os.path.join(self.expath(cachepath, uuid, idtype),uuid+'.json')

    def artexpath(self, uuid, cachepath, arttype):
        return os.path.join(cachepath,'artist',self.upath(uuid),arttype)

    def country(self, code):
        if code == 'XW':
            return "worldwide"
        elif code == 'XE':
            return "across Europe"
        else:
            c=pycountry.countries.get(alpha_2=code)
            if c.name:
                return "in "+c.name
            else:
                return "in "+code
                
    def mkdir_p(self, path):
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise
    
    def trim_image(self, filename):
        image = PImage.open(filename)
        try:
            # convert to RGBa before getting bounding box to account for
            # transparent pixels
            bbox = image.convert("RGBa").getbbox()
        except ValueError:
            # if grayscale, do some fancy compositing
            oimage = image.convert('RGBA')
            bimage = PImage.new('RGBA',image.size)
            bbox = PImage.composite(oimage,bimage,oimage).getbbox()
        # crop it and save
        image = image.crop(bbox)
        image.save(filename)



import sys
import os
import argparse
import json
import musicbrainzngs
import re
import errno
import io
import urllib
from twisted.internet import reactor, protocol, threads, defer, task
from twisted.web import client, http, error as weberror
from twisted.python import log
from mpd import MPDProtocol
from functools import partial
from kmpc.mpdfactory import MpdConnection
from kmpc.extra import KmpcHelpers
from kmpc.version import VERSION_STR

Helpers = KmpcHelpers()
client.HTTPClientFactory.noisy = False

class MpdParser(object):
    """The main class which queries mpd and dispatches events."""
    def __init__(self, args):
        self.mode = []
        # check what modes were specified, if none then do everything
        if (not args.artists and not args.releases and not args.recordings
                and not args.symlinks and not args.artistart):
            self.mode = ['artists','releases',
                    'recordings','symlinks','artistart']
        else:
            if args.artists:
                self.mode.append('artists')
            if args.releases:
                self.mode.append('releases')
            if args.recordings:
                self.mode.append('recordings')
            if args.symlinks:
                self.mode.append('symlinks')
            if args.artistart:
                self.mode.append('artistart')
        # these variables track what uuids we need to operate on
        self.queue = {}
        self.queue['artist'] = {}
        self.queue['release'] = {}
        self.queue['recording'] = {}
        self.queue['artistsymlink'] = {}
        self.queue['artistart'] = {}
        self.args = args
        # seconds to wait before next http query (so we don't get blocked)
        self.waittime = 0
        # how many tasks are left to wait on
        self.waitleft = 0
        # how many mpd queries are left to wait on
        self.mpdqueries = 0
        log.startLogging(sys.stdout)
        log.msg("Starting MpdParser with mode "+format(self.mode))
        musicbrainzngs.set_useragent(
                "kmpc",
                VERSION_STR,
                'https://gitlab.com/eratosthene/kmpc')
        self.mpdconnection = MpdConnection(
                args.ip, args.port,
                None, [self.dispatch], True, False)

    def dispatch(self, conn):
        """Start the callback chain for each mode."""
        if 'artists' in self.mode:
            Helpers.mkdir_p(os.path.join(self.args.cachepath,'artist'))
            # if --artistid specified, operate only on that list
            if self.args.artistid:
                d = defer.Deferred()
                d.addCallback(partial(self.handle_list,'artist'))
                d.addErrback(log.err)
                d.callback(self.args.artistid.split(','))
            else:
                (self.mpdconnection.protocol.
                        list('musicbrainz_artistid').
                        addCallback(partial(self.handle_list,'artist')).
                        addErrback(self.mpdconnection.handle_mpd_error))
                self.mpdqueries += 1
        if 'releases' in self.mode:
            Helpers.mkdir_p(os.path.join(self.args.cachepath,'release'))
            # if --releaseid specified, operate only on that list
            if self.args.releaseid:
                d = defer.Deferred()
                d.addCallback(partial(self.handle_list,'release'))
                d.addErrback(log.err)
                d.callback(self.args.releaseid.split(','))
            else:
                (self.mpdconnection.protocol.
                        list('musicbrainz_albumid').
                        addCallback(partial(self.handle_list,'release')).
                        addErrback(self.mpdconnection.handle_mpd_error))
                self.mpdqueries += 1
        if 'recordings' in self.mode:
            Helpers.mkdir_p(os.path.join(self.args.cachepath,'recording'))
            # if --recordingid specified, operate only on that list
            if self.args.recordingid:
                d = defer.Deferred()
                d.addCallback(partial(self.handle_list,'recording'))
                d.addErrback(log.err)
                d.callback(self.args.recordingid.split(','))
            else:
                (self.mpdconnection.protocol.
                        list('musicbrainz_trackid').
                        addCallback(partial(self.handle_list,'recording')).
                        addErrback(self.mpdconnection.handle_mpd_error))
                self.mpdqueries += 1
        if 'symlinks' in self.mode:
            Helpers.mkdir_p(os.path.join(self.args.cachepath,'artist_by_name'))
            # if --artistid specified, operate only on that list
            if self.args.artistid:
                d = defer.Deferred()
                d.addCallback(self.handle_symlink_list)
                d.addErrback(log.err)
                d.callback(self.args.artistid.split(','))
            else:
                (self.mpdconnection.protocol.
                        list('musicbrainz_artistid').
                        addCallback(self.handle_symlink_list).
                        addErrback(self.mpdconnection.handle_mpd_error))
                self.mpdqueries += 1
            self.waitleft += 1
        if 'artistart' in self.mode:
            Helpers.mkdir_p(os.path.join(self.args.cachepath,'artist'))
            # if --artistid specified, operate only on that list
            if self.args.artistid:
                d = defer.Deferred()
                d.addCallback(self.handle_art_list)
                d.addErrback(log.err)
                d.callback(self.args.artistid.split(','))
            else:
                (self.mpdconnection.protocol.
                        list('musicbrainz_artistid').
                        addCallback(self.handle_art_list).
                        addErrback(self.mpdconnection.handle_mpd_error))
                self.mpdqueries += 1

    def check_exit(self, *args):
        """Decrements the task counter, then exits if 0."""
        self.waitleft -= 1;
        log.msg("check_exit: "+str(self.waitleft)+" tasks left")
        if self.waitleft < 1:
            reactor.stop()

    def handle_list(self, qtype, result):
        """Parses the results of an mpd query and acts on it."""
        for row in result:
            # MB tags can have more than one uuid separated by /
            for uuid in str(row).split('/'):
                if len(uuid) > 0 and uuid not in self.queue[qtype]:
                    checkpath = Helpers.jpath(self.args.cachepath, uuid, qtype)
                    # only query if file doesn't exist or we are ignoring it
                    if self.args.ignorecache or not os.path.isfile(checkpath):
                        # add uuid to the queue so it only gets scheduled once
                        self.queue[qtype][uuid] = 1
                        log.msg("Scheduling MB query for "+qtype+" "+uuid+
                                " in "+str(self.waittime)+" seconds")
                        reactor.callLater(self.waittime, self.query_mb,
                                uuid, qtype)
                        self.waittime += 1
                        self.waitleft += 1
        # disconnect mpd after all queries are complete
        self.mpdqueries -= 1
        if self.mpdqueries < 1:
            self.mpdconnection.reactor.disconnect()
        # if we got all the way here and didn't do anything, we're done
        if self.waitleft == 0:
            self.check_exit()

    def handle_symlink_list(self, result):
        """Creates symlinks from an artist's readable name to their artistid."""
        for row in result:
            # MB tags can have more than one uuid separated by /
            for uuid in str(row).split('/'):
                checkpath = Helpers.jpath(self.args.cachepath, uuid, 'artist')
                if len(uuid) > 0 and os.path.isfile(checkpath):
                    # open the artist json file and parse it
                    with open(checkpath, 'r') as infile:
                        data = json.load(infile)
                        # create a UTF-8, os-safe name/sort_name to symlink to
                        name = (u''.join(data['artist']['name'])
                                .encode('utf-8').strip().replace(os.sep, '_'))
                        sort_name = (u''.join(data['artist']['sort-name'])
                                .encode('utf-8').strip().replace(os.sep, '_'))
                        if uuid not in self.queue['artistsymlink']:
                            # add uuid to the queue so it only gets scheduled
                            # once
                            self.queue['artistsymlink'][uuid] = 1
                            log.msg(u"Creating symlink for ["+
                                    name+"] ["+sort_name+"]")
                            try:
                                # symlink to name, ignoring errors
                                os.symlink(
                                        os.path.join('..', 
                                                'artist',
                                                Helpers.upath(uuid)),
                                        os.path.join(self.args.cachepath,
                                                'artist_by_name',
                                                name))
                            except OSError:
                                pass
                            try:
                                # symlink to sort_name, ignoring errors
                                os.symlink(
                                        os.path.join('..',
                                                'artist',
                                                Helpers.upath(uuid)),
                                        os.path.join(self.args.cachepath,
                                                'artist_by_name',
                                                sort_name))
                            except OSError:
                                pass
        # disconnect mpd after all queries are complete
        self.mpdqueries -= 1
        if self.mpdqueries < 1:
            self.mpdconnection.reactor.disconnect()
        self.check_exit()

    def query_mb(self, uuid, qtype, *args):
        """Query musicbrainz in a separate thread."""
        if qtype == 'artist':
            d = threads.deferToThread(musicbrainzngs.get_artist_by_id, uuid)
            d.addCallback(partial(self.handle_query_mb_artist, uuid))
        elif qtype == 'release':
            d = threads.deferToThread(musicbrainzngs.get_release_by_id, uuid,
                    includes=["recordings","release-groups","artists"])
            d.addCallback(partial(self.handle_query_mb_release, uuid))
        elif qtype == 'recording':
            d = threads.deferToThread(musicbrainzngs.get_recording_by_id, uuid, 
                    includes=["artists","releases"])
            d.addCallback(partial(self.handle_query_mb_recording, uuid))
        self.waittime -= 1

    def handle_query_mb_artist(self, artist_id, mbres):
        """Write the artist data to a json file."""
        name = u''.join(mbres['artist']['name']).encode('utf-8').strip()
        sort_name = (u''.join(mbres['artist']['sort-name'])
                .encode('utf-8').strip())
        # TODO: for some reason we still get ascii exceptions sometimes
        log.msg(u"Found artist ["+name+"] ["+
                sort_name+"]")
        Helpers.mkdir_p(Helpers.expath(self.args.cachepath,
                artist_id, 'artist'))
        with open(Helpers.jpath(self.args.cachepath,
                artist_id, 'artist'), 'w') as outfile:
            json.dump(mbres, outfile)
        self.check_exit()

    def handle_query_mb_release(self, release_id, mbres):
        """Write the release data to a json file."""
        title = u''.join(mbres['release']['title']).encode('utf-8').strip()
        # TODO: for some reason we still get ascii exceptions sometimes
        log.msg(u"Found release ["+title+"] "+str(self.waitleft-1)+"}")
        Helpers.mkdir_p(Helpers.expath(self.args.cachepath,
                release_id, 'release'))
        with open(Helpers.jpath(self.args.cachepath,
                release_id, 'release'), 'w') as outfile:
            json.dump(mbres, outfile)
        self.check_exit()

    def handle_query_mb_recording(self, recording_id, mbres):
        """Write the recording data to a json file."""
        title = u''.join(mbres['recording']['title']).encode('utf-8').strip()
        # TODO: for some reason we still get ascii exceptions sometimes
        log.msg(u"Found recording ["+title+"] "+str(self.waitleft-1)+"}")
        Helpers.mkdir_p(Helpers.expath(self.args.cachepath,
                recording_id, 'recording'))
        with open(Helpers.jpath(self.args.cachepath,
                recording_id, 'recording'), 'w') as outfile:
            json.dump(mbres, outfile)
        self.check_exit()

    def handle_art_list(self, result):
        """Download artwork for all artists."""
        self.mpdqueries -= 1
        if self.mpdqueries < 1:
            self.mpdconnection.reactor.disconnect()
        for row in result:
            for uuid in str(row).split('/'):
                if len(uuid) > 0:
                    # add to queue so it only gets scheduled once
                    self.queue['artistart'][uuid] = 1
        artists = self.queue['artistart'].keys()
        artists.sort()
        log.msg("Downloading art for "+str(len(artists))+" artists")
        fanarturl = "http://webservice.fanart.tv/v3/music/"
        apikey = "406b2a5af85c14b819c1c6332354b313"
        for uuid in artists:
            self.waitleft += 1
            furl = fanarturl+uuid+"?api_key="+apikey
            log.msg("scheduling fanart.tv query for "+uuid+" in "+
                    str(self.waittime)+" seconds")
            d = task.deferLater(reactor, self.waittime, client.getPage, furl)
            self.waittime += 1
            d.addCallback(partial(self.handle_agent_response, uuid))
            # trap 404 errors so they don't get displayed
            d.addErrback(self.ignore_404).addErrback(log.err)
        # if we got all the way here and didn't do anything, we're done
        if self.waitleft == 0:
            self.check_exit()

    def ignore_404(self, f):
        """Ignores 404 errors and decrements the task counter."""
        r = f.trap(weberror.Error)
        self.waittime -= 1
        self.check_exit()
        if f.value.status != '404':
            return f

    def handle_agent_response(self, uuid, body):
        """Parses json response from fanart.tv."""
        d = json.loads(body)
        log.msg(u"Found artist "+d['name'])
        if 'hdmusiclogo' in d:
            for idx, img in enumerate(d['hdmusiclogo']):
                self.check_file_download(uuid, img,
                        'logo', 'png')
        if 'musiclogo' in d:
            for idx, img in enumerate(d['musiclogo']):
                self.check_file_download(uuid, img,
                        'logo', 'png')
        if 'artistbackground' in d:
            for idx, img in enumerate(d['artistbackground']):
                self.check_file_download(uuid, img,
                        'artistbackground', 'jpg')
        self.waittime -= 1
        self.check_exit()

    def handle_download_file_response(self, filepath, trim, body):
        """Writes downloaded image to disk."""
        log.msg("Writing to "+filepath)
        file = open(filepath, "wb")
        file.write(body)
        file.close()
        # if this is an artist logo, trim any extra transparent space
        if trim:
            log.msg("trimming file")
            Helpers.trim_image(filepath)
        self.waittime -= 1
        self.check_exit()

    def check_file_download(self, uuid, img, arttype, ext):
        """Checks for an existing image file, downloading it if necessary."""
        fp = os.path.join(Helpers.artexpath(uuid, self.args.cachepath, arttype),
                img['id']+"."+ext)
        log.msg("Checking for "+fp)
        self.waitleft += 1
        Helpers.mkdir_p(Helpers.artexpath(uuid, self.args.cachepath, arttype))
        # only download if file doesn't exist or we are ignoring cache
        if self.args.ignorecache or not os.path.isfile(fp):
            log.msg("Scheduling download of "+arttype+" "+img['id']+
                    " in "+str(self.waittime)+" seconds")
            log.msg("from "+str(img['url']))
            d = task.deferLater(reactor, self.waittime, client.getPage,
                    str(img['url']))
            self.waittime += 1
            d.addCallback(partial(self.handle_download_file_response,
                    fp, (arttype == 'logo')))
            # trap 404 errors so they don't get displayed
            d.addErrback(self.ignore_404).addErrback(log.err)


def main_app():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--ip",
            help="ip address or hostname of mpd server (default 127.0.0.1)",
            default="127.0.0.1")
    parser.add_argument("-p", "--port",
            help="port of mpd server (default 6600)",
            default="6600")
    parser.add_argument("-c", "--cachepath",
            help="directory to save mb cach (default ~/.kmpc/mb-cache)",
            default=os.path.expanduser('~/.kmpc/mb-cache'))
    parser.add_argument("--artists",
            help="scan artists", action="store_true")
    parser.add_argument("--releases",
            help="scan releases", action="store_true")
    parser.add_argument("--recordings",
            help="scan recordings", action="store_true")
    parser.add_argument("--symlinks",
            help="create artist symlinks", action="store_true")
    parser.add_argument("--artistart",
            help="download artist backgrounds and logos", action="store_true")
    parser.add_argument("--ignorecache",
            help="ignore existing cache, rescan everything",
            action="store_true")
    parser.add_argument("--artistid",
            help="comma-delimited list of artistids to operate on")
    parser.add_argument("--releaseid",
            help="comma-delimited list of releaseids to operate on")
    parser.add_argument("--recordingid",
            help="comma-delimited list of recordingids to operate on")
    args = parser.parse_args()
    print format(args)
    mpdparser = MpdParser(args)
    reactor.run()

if __name__ == "__main__":
    main_app()



from functools import partial

import kivy
from kivy.logger import Logger
from kivy.uix.screenmanager import ScreenManager, CardTransition
from kivy.app import App
from kivy.support import install_twisted_reactor
from kivy.properties import ObjectProperty

from kmpc.nowplaying import NowPlayingScreen
from kmpc.queue import QueueScreen
from kmpc.library import LibraryScreen
from kmpc.system import SystemScreen
from kmpc.mpdfactory import MpdConnection

# make sure we are on updated version of kivy
kivy.require('1.10.0')

class KmpcScreenManager(ScreenManager):
    """The main class that ties it all together."""

    mpdconnection = ObjectProperty(None)
    syncmpdconnection = ObjectProperty(None)
    nowplaying = ObjectProperty(None)
    queue = ObjectProperty(None)
    library = ObjectProperty(None)
    system = ObjectProperty(None)

    def __init__(self, config, *args, **kwargs):
        """Zero out variables, pull in config file, connect to mpd."""
        super(KmpcScreenManager, self).__init__(*args, **kwargs)
        self.transition = CardTransition(direction='down')
        self.config = config
        self.nowplaying = NowPlayingScreen()
        self.nowplaying.bind(on_enter=self.nowplaying.enter_screen)
        self.nowplaying.bind(on_leave=self.nowplaying.leave_screen)
        self.add_widget(self.nowplaying)
        self.queue = QueueScreen()
        self.queue.bind(on_enter=self.queue.enter_screen)
        self.add_widget(self.queue)
        self.library = LibraryScreen()
        self.library.bind(on_enter=self.library.enter_screen)
        self.add_widget(self.library)
        self.system = SystemScreen()
        self.system.bind(on_enter=self.system.enter_screen)
        self.add_widget(self.system)
        self.mpd_status = {'state': 'stop',
                           'repeat': 0,
                           'single': 0,
                           'random': 0,
                           'consume': 0,
                           'curpos': 0}
        self.currsong = None
        self.nextsong = None
        self.currfile = None
        self.do_idle_handler = True
        # install twisted reactor to interface with mpd
        install_twisted_reactor()
        # open mpd connection
        self.mpdconnection = MpdConnection(
                self.config.get('mpd', 'mpdhost'),
                self.config.get('mpd', 'mpdport'),
                self.mpd_idle_handler,
                [self.init_mpd],
                Logger.info)

    def init_mpd(self, instance):
        # get the initial mpd status
        (self.mpdconnection.protocol.status().
            addCallback(self.nowplaying.update_mpd_status).
            addErrback(self.mpdconnection.handle_mpd_error))
        # subscribe to 'kmpc' to check for messages from mpd
        self.mpdconnection.protocol.subscribe('kmpc')

    def mpd_idle_handler(self, result):
        # global flag for disabling during sync
        if self.do_idle_handler:
            # notify various subsystems based on what changed
            for s in result:
                Logger.info('mpd_idle_handler: Changed '+format(s))
                if format(s) == 'playlist':
                    # queue was changed, ask mpd for playlist info
                    (self.mpdconnection.protocol.playlistinfo().
                        addCallback(self.queue.populate_queue).
                        addErrback(self.mpdconnection.handle_mpd_error))
                    # force a reload of nextsong if queue changes
                    self.nextsong = None
                    (self.mpdconnection.protocol.status().
                        addCallback(self.nowplaying.update_mpd_status).
                        addErrback(self.mpdconnection.handle_mpd_error))
                elif format(s) == 'player':
                    # player was changed, ask mpd for player status
                    (self.mpdconnection.protocol.status().
                        addCallback(self.nowplaying.update_mpd_status).
                        addErrback(self.mpdconnection.handle_mpd_error))
                elif format(s) == 'sticker':
                    # song rating sticker was changed, ask mpd for current song
                    # rating
                    (self.mpdconnection.protocol.status().
                        addCallback(self.nowplaying.update_mpd_status).
                        addErrback(self.mpdconnection.handle_mpd_error))
                    (self.mpdconnection.protocol.
                        sticker_get('song', self.currfile, 'rating').
                        addCallback(self.nowplaying.update_mpd_sticker_rating).
                        addErrback(self.mpdconnection.handle_mpd_no_sticker))
                elif format(s) == 'options':
                    # some playback option was changed, ask mpd for player
                    # status
                    (self.mpdconnection.protocol.status().
                        addCallback(self.nowplaying.update_mpd_status).
                        addErrback(self.mpdconnection.handle_mpd_error))
                elif format(s) == 'message':
                    # an mpd message was received, ask mpd what it was
                    (self.mpdconnection.protocol.readmessages().
                        addCallback(self.handle_mpd_message).
                        addErrback(self.mpdconnecion.handle_mpd_error))
                else:
                    # default if none of the above, ask mpd for player status
                    (self.mpdconnection.protocol.status().
                        addCallback(self.nowplaying.update_mpd_status).
                        addErrback(self.mpdconnection.handle_mpd_error))

    def change_replaygain(self, v):
        Logger.debug("NowPlaying: change_replaygain to " + format(v))
        self.mpdconnection.protocol.replay_gain_mode(
                str(v)).addErrback(self.mpdconnection.handle_mpd_error)

    def change_crossfade(self, v):
        """Callback when user changes crossfade slider."""
        Logger.info('Settings: change_crossfade')
        self.mpdconnection.protocol.crossfade(
                str(v)).addErrback(self.mpdconnection.handle_mpd_error)

    def change_mixrampdb(self, v):
        """Callback when user changes mixrampdb slider."""
        Logger.info('Settings: change_mixrampdb')
        self.mpdconnection.protocol.mixrampdb(
                str(0.0-v)).addErrback(self.mpdconnection.handle_mpd_error)

    def change_mixrampdelay(self, v):
        """Callback when user changes mixrampdelay slider."""
        Logger.info('Settings: change_mixrampdelay')
        self.mpdconnection.protocol.mixrampdelay(
                str(v)).addErrback(self.mpdconnection.handle_mpd_error)




import sys
import argparse
from twisted.internet import reactor,protocol
from twisted.python import log
from mpd import MPDProtocol
from kmpc.mpdfactory import MpdConnection

class MessageHandler(object):
    """Handles mpd messages."""
    def __init__(self, mpdconnection, error_handler):
        self.mpdconnection = mpdconnection
        self.handle_mpd_error = error_handler
        self.uri = None
        self.rating = None
        self.rid = None

    def handle_mpd_message(self, result):
        """Gets current rating and recordingid for a song."""
        for m in result:
            self.uri= m['message']
            log.msg("Rating changed for "+self.uri)
            (self.mpdconnection.protocol.sticker_get('song',self.uri,'rating').
                    addCallback(self.handle_sticker_get).
                    addErrback(self.handle_mpd_error))
            (self.mpdconnection.protocol.lsinfo(self.uri).
                    addCallback(self.handle_lsinfo).
                    addErrback(self.handle_mpd_error))

    def handle_lsinfo(self, result):
        """Finds all songs with the same recordingid."""
        for r in result:
            if 'musicbrainz_trackid' in r:
                self.rid=r['musicbrainz_trackid']
                log.msg('found recording id '+self.rid)
                (self.mpdconnection.protocol.find('MUSICBRAINZ_TRACKID',self.rid).
                        addCallback(self.handle_find).
                        addErrback(self.handle_mpd_error))

    def handle_find(self, result):
        """Sets rating for a song if it has the same recordingid."""
        for r in result:
            f = r['file']
            if f != self.uri:
                if self.rating is not None:
                    log.msg("Setting rating for "+f+" to "+self.rating)
                    self.mpdconnection.protocol.sticker_set('song',f,'rating',self.rating)
                else:
                    log.msg("Rating not yet set, debug race condition")

    def handle_sticker_get(self, result):
        """Stores rating for a song."""
        log.msg("found rating "+format(result))
        self.rating = format(result)

class MpdRatingSync(object):
    """Main application class."""
    def __init__(self, args):
        log.startLogging(sys.stdout)
        self.mpdconnection = MpdConnection(
                args.ip, args.port,
                self.mpd_idle_handler,
                [self.subscribe_to_channel],
                True,True)

    def mpd_idle_handler(self, result):
        """Dispatches MessageHandler on incoming messages."""
        for s in result:
            if format(s) == 'message':
                log.msg('Event on subsystem '+format(s))
                mh = MessageHandler(self.mpdconnection,
                        self.mpdconnection.handle_mpd_error)
                (self.mpdconnection.protocol.readmessages().
                        addCallback(mh.handle_mpd_message).
                        addErrback(self.mpdconnection.handle_mpd_error))

    def subscribe_to_channel(self, conn):
        """Subscribes to ratingchange channel."""
        log.msg("subscribing to ratingchange")
        self.mpdconnection.protocol.subscribe("ratingchange")

def main_app():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--ip",
            help="ip address or hostname of mpd server (default 127.0.0.1)",
            default='127.0.0.1')
    parser.add_argument("-p", "--port",
            help="port of mpd server (default 6600)",
            default='6600')
    args = parser.parse_args()
    mpdratingsync = MpdRatingSync(args)
    reactor.run()

if __name__ == "__main__":
    main_app()